/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XC40 object specializes the XCPORT object so that the XC4000 FPGA on the
// XS40 Board can be configured and the RAM can be loaded.  The XC40 object controls
// the JTAG port of the XC4000 chip, and also the /PROGRAM, DIN, and CCLK pins.
//

// $Log: xc40.cpp $
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:00:48  DEVB
// Initial revision
//

#include <fstream.h>
#include "xcbsdr.h"
#include "xc40.h"
#include "os.h"

// The XC4000 does not have built-in JTAG circuitry that we can use to access the
// RAM chip on the XS40 Board.  So we have to load a pseudo-JTAG circuit into the
// XC4000 so we can load data into the RAM.  The following values describe the
// boundary scan data and instruction registers of the pseudo-JTAG circuitry.
 
// pseudo-JTAG has no instruction register but we set it to be 1 bit anyway
const unsigned int bsirLength = 1;
// 26 pins needed to control RAM, so pseudo-JTAG BSDR has 3*26 bits
const unsigned int bsdrLength = 3 * 26;

// position of cells in the BSDR which affect the address bits of the
// RAM on the XS40 Board
static unsigned int xc40RamAddressBitPos[numRAMAddressBits] =
{	//A0 .... A14
	0,3,6,9,12,15,18,21,24,27,30,33,36,39,42,
};

// position of the cells in the BSDR which affect the data bits of the
// RAM on the XS40 Board
static unsigned int xc40RamDataBitPos[numRAMDataBits] =
{	//D0 .... D7
	45,48,51,54,57,60,63,66,
};

// position of the cells in the BSDR which affect the CS, OE, and WE
// bits of the RAM on the XS40 board
static unsigned int xc40RamControlBitPos[numRAMControlBits] =
{	// CS,OE,WE
	69,72,75,
};


// begin methods and functions

// Create an XC40 object.  The type of XC4000 FPGA on the XS40 Board is inferred
// by reading the chip type from a bitfile that is passed to the constructor.
// The directory containing the pseudo-JTAG bitstreams for all the different types
// of XC4000 chips is also passed as an argument.
XC40::XC40(XSError* e,	// error reporting channel
		   unsigned int portNum,	// parallel port number
		   string& bitfileName,		// bitfile with XC4000 configuration bitstream
		   string& loadDir			// directory where pseudo-JTAG bitstreams are stored
		   ) : XCPort(e,portNum)	// instantiate the XCPort base object
{	
	// make sure the PROGRAM pin is high
	SetPROGRAM(1);
	// pull all the data pins high to keep XS Board microcontroller in reset state
	// (This is only needed if the bitstream programs the FPGA so that one of
	// the parallel port pins drives the reset of the microcontroller.)
	SetData(0xff);
	// get the chip identifier from the bitstream file
	SetChipType(bitfileName);
	// store the directory where the RAM interface pseudo-JTAG bitstreams are stored
	ramLoaderDir = loadDir;
	ramLoaderIsRdy = false;	// chip has not yet been loaded with the RAM interface bitstream
}


// bit pos of /PROG pin in par. port control reg.
static const unsigned int posPROG = 0;

// Set the value on the PROGRAM pin of the XC4000 FPGA
void XC40::SetPROGRAM(unsigned int b)
{
	SetControl(b,posPROG,posPROG);
}


// Initialize the XC4000 FPGA for configuration with DIN, CCLK, and PROGRAM pins
void XC40::InitConfigureFPLD(void)
{
	SetPROGRAM(1); // pulse the PROGRAM pin to set the mode
	SetPROGRAM(0); //   and begin configuration of the FPGA
	SetPROGRAM(1);
	
	// insert 300 ms delay after PROGRAM pin goes high
	InsertDelay(300,MILLISECONDS);
}


// Send out a byte of configuration information
void XC40::ConfigureFPLD(unsigned char b)
{
	// configuration bytes for the XC4000 programming pins only contain
	// values for the DIN pin, so eight clock pulses are needed per byte.
	for(int i=0; i<8; i++)
	{
		SetTDI(b&0x80 ? 1:0);  // byte is sent MSB first
		b <<= 1;
		PulseTCK();
	}
}


// compute an integer from a vector of hex numbers in a stream
static long unsigned int GetInteger(istream& is, int len)
{
	assert(len<=4); // can't handle lengths greater than 32 bits
	long unsigned int lui = 0;
	for(int i=0; i<len; i++)
	{ // process hex digits starting from MSDigit
		unsigned char digit;
		is.read(&digit,1);
		assert(is.eof()==0);
		lui = lui*256 + (long unsigned int)digit;
	}
	return lui;
}

// get the field type indicator from a stream
static int GetType(istream& is)
{
	unsigned char type;
	is.read(&type,1);
	assert(is.eof()==0);
	return type;
}

// pass over fields until a certain field is found
static bool ScanForField(istream& is, unsigned char searchType)
{
	while(true)
	{
		unsigned char type = GetType(is);
		if(type==searchType)
			return true;
		unsigned int length = GetInteger(is);
		is.ignore(length);
		if(is.eof()!=0)
			break;
	}
	return false;
}

const int BitstreamFieldType = 0x65;	// field type for the bitstream data

// process a stream and send the configuration bitstream to the board
bool XC40::ConfigureFPLD(istream& is)
{
	XSError& err = GetErr();
	long unsigned int i;
	
	if(is.eof()!=0)
		return false; // exit if no BIT stream exists
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is);
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is);
	assert(fieldLength==1);
	
	// now look for the field with the bitstream data
	bool status = ScanForField(is,BitstreamFieldType);
	assert(status==true);
	
	// now we are at the start of the configuration bits
	fieldLength = GetInteger(is,4); // get the length of the bit stream
	assert(fieldLength>0);
#if DEBUG
	cerr << "bit stream field length = " << fieldLength << endl;
#endif
	InitConfigureFPLD(); // start the configuration process
	for(i=0; i<fieldLength; i++)
	{ // read the bytes and send the config. bits to the board
		unsigned char byte;
		is.read(&byte,1);
		assert(is.eof()==0);	// should not hit end-of-file
		ConfigureFPLD(byte);
		// output some feedback as the configuration proceeds
		if((i%1000)==0)
			cerr << "*";
	}
	cerr << endl;
	
	// clock a few more times at the end of the configuration process
	// to clean things up...
	for(i=0; i<4; i++)
		PulseTCK();
	
#if DEBUG
	cerr << "exiting ConfigureFPLD()" << endl;
#endif
	
	return true;
}

const int DeviceFieldType = 0x62;		// field type for the FPGA device identifier

// gets the chip identifier from the bitstream and stores it in the object
bool XC40::SetChipType(istream& is)
{
	XSError& err = GetErr();
	long unsigned int i;
	
	if(is.eof()!=0)
		return false; // exit if no BIT stream exists
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is);
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is);
	assert(fieldLength==1);
	
	// now look for the field with the chip identifier
	bool status = ScanForField(is,DeviceFieldType);
	assert(status==true);
	fieldLength = GetInteger(is);
	assert(fieldLength>0);
	char buf[8];
	for(i=0; i<7 && i<fieldLength; i++)
		is >> buf[i];
	buf[i] = 0;		// terminate string
	chipType = buf;	// store the chip identifier
	return true;	// everything went OK
}

// gets the chip identifier from the bitstream and stores it in the object
bool XC40::SetChipType(string& bitfileName)
{
	XSError& err = GetErr();
	
	if(bitfileName.compare("")==0)
		return false; // exit if no BIT file was given
	
	ifstream is(bitfileName.data(),ios::binary|ios::nocreate);
	if(is.fail() || is.eof())
	{ // error if given file could not be opened
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << bitfileName.data() << endl;
		err.EndMsg();
	}
	
	// now get the chip identifier from the bitstream file
	bool status = SetChipType(is);
	
	is.close();
	
	return status;
}

// configures an FPGA so the XS Board RAM can be loaded
bool XC40::ConfigureForRAMDownload()
{
	XSError& err = GetErr();
	
	// create the JTAG registers that will interface to the XS40 RAM
	bsirPtr = new Bitstream(bsirLength);
	bsdrPtr = new XCBSDR(bsdrLength,
		xc40RamAddressBitPos,
		xc40RamDataBitPos,
		xc40RamControlBitPos);

	string loadingCircuitFileName = ramLoaderDir;
	loadingCircuitFileName += DIRSEP;
	loadingCircuitFileName += XC40PREFIX;
	loadingCircuitFileName += chipType;
	loadingCircuitFileName += XC40SUFFIX;
	return ramLoaderIsRdy = XCPort::ConfigureFPLD(loadingCircuitFileName);
}

// Send a byte into an Atmel EEPROM
void XC40::SendEEPROMByte(unsigned char byte)
{
	assert(GetTCK()==0); // TCK should be zero on entry
	for(int i=0; i<8; i++)
	{
		// TDI pin of parallel port connects to data pin of Atmel EEPROM
		SetTDI(byte&0x80 ? 1:0);  // byte is sent MSB first
		byte <<= 1;
		// TCK pin of parallel port connects to CLK pin of Atmel EEPROM
		PulseTCK();
	}
	SetTDI(1);	// this is for the acknowledge bit from the EEPROM
	PulseTCK();
	assert(GetTCK()==0); // TCK should be zero on exit
}

// Program a page of an Atmel EEPROM
void XC40::ProgramEEPROMPage(unsigned char* buf,	// buffer with data bytes
							 unsigned int pageAddr,	// address where data will be stored
							 unsigned int len		// number of bytes in buffer
							 )
{
	assert(len<=64);	// an Atmel page is 64 bytes or less
	assert(pageAddr<0x20000L);	// Atmel byte addresses are less than 128K
	
	// initialize TCK and TDI
	SetTCK(0);
	SetTDI(1);
	// give start signal (TDI falls while TCK is high)
	SetTCK(1);
	SetTDI(0);
	SetTCK(0);
	// send device-address byte with A2=0 and R/W=0 (write enabled)
	SendEEPROMByte(0xA6);
	// send two-byte address of EEPROM page
	SendEEPROMByte(pageAddr>>8);	// send upper byte
	SendEEPROMByte(pageAddr&0xFF);	// send lower bytes
	// send data bytes in page
	for(unsigned int i=0; i<len; i++)
		SendEEPROMByte(buf[i]);
	// give stop signal (TDI rises while TCK is high)
	SetTDI(0);
	SetTCK(1);
	SetTDI(1);
	// wait 10 ms
	InsertDelay(10,MILLISECONDS);
}

// program an Atmel serial EEPROM with a bitstream
bool XC40::ProgramEEPROM(istream& is)
{
	XSError& err = GetErr();
	long unsigned int i;

	if(is.eof()!=0)
		return false; // exit if no BIT stream exists

	// We need to load the pseudo-JTAG circuit that is used to load the RAM
	// because it allows us to access the pins that are used to program
	// the serial EEPROM on the XS40 Board.
	if(ConfigureForRAMDownload()==false)
		return false;

	// Setup the pins of the Atmel EEPROM for programming.
	// The SER_EN_ should already be held low by a shunt on Jumper J6.
	// This holds the EEPROM CE_ low.
	SetData(0);
	// The following uses the JTAG circuitry to lower the RESET_ input.
	XCBSDR& bsdr = *bsdrPtr;
	Bitstream null(0);
	bsdr.Clear();
	bsdr.SetRAMData(0x00); // this will clear RESET_
	InitTAP();  // reset the TAP state controller
	GoThruTAPSequence(RunTestIdle,SelectDRScan,CaptureDR,ShiftDR, -1);
	SendRcvBitstream(bsdr,null);
	GoThruTAPSequence(UpdateDR,-1); // this sets the EEPROM RESET low
	InitTAP();  // reset the TAP state controller
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is);
#if DEBUG
	cerr << "first field length = " << fieldLength << endl;
#endif
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is);
#if DEBUG
	cerr << "second field length = " << fieldLength << endl;
#endif
	assert(fieldLength==1);
	
	// now look for the field with the bitstream data
	bool status = ScanForField(is,BitstreamFieldType);
	assert(status==true);
	
	// now we are at the start of the configuration bits
	fieldLength = GetInteger(is,4); // get the length of the bit stream
	assert(fieldLength>0);
#if DEBUG
	cerr << "bit stream field length = " << fieldLength << endl;
#endif
	
	// now load the EEPROM memory
	unsigned char page[64];
	for(long unsigned int i=0; i<fieldLength; i++)
	{ 
		is.read(&page[i&0x3f],1);	// read the bytes and into a 64-byte page
		if(is.eof())
		{
			is.eof();
		}
		assert(is.eof()==0);	// shouldn't reach end-of-file
		if((i%1000)==0)		// after every 1000 bytes,
			cerr << "@";	// give some feedback during the programming process
		if((i&0x3f)==0x3f)
		{ // got a full 64-byte page, so program the EEPROM
			ProgramEEPROMPage(page,i&~0x3F,64);
		}
	}
	if((i&0x3F)!=0)
	{ // BIT file ended before a full 64-byte page was assembled
		ProgramEEPROMPage(page,i&~0x3F,(i&0x3F)+1);
	}
	cerr << endl;	// end the string of @ symbols
	
	return true;
}
