/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XC40 object specializes the XCPORT object so that the XC4000 FPGA on the
// XS40 Board can be configured and the RAM can be loaded.  The XC40 object controls
// the JTAG port of the XC4000 chip, and also the /PROGRAM, DIN, and CCLK pins.
//

// $Log: xc40.h $
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:57:56  DEVB
// Initial revision
//

#ifndef XC40_H
#define XC40_H

#include "xcport.h"

class XC40 : public XCPort
{
	public:

	/* begin XC40 methods */

	/* method: XC40 */
	// Create an XC40 object.  The type of XC4000 FPGA on the XS40 Board is inferred
	// by reading the chip type from a bitfile that is passed to the constructor.
	// The directory containing the pseudo-JTAG bitstreams for all the different types
	// of XC4000 chips is also passed as an argument.
	 XC40(XSError* e, // error reporting channel
		unsigned int portNum, // parallel port number
		string& bitfileName, // bitfile with XC4000 configuration bitstream
		string& loadDir // directory where pseudo-JTAG bitstreams are stored
		);

	/* method: ConfigureFPLD */
	// Send out a byte of configuration information
	void ConfigureFPLD(unsigned char b);

	/* method: ConfigureForRAMDownload */
	// configures an FPGA so the XS Board RAM can be loaded
	bool ConfigureForRAMDownload();

	/* method: ProgramEEPROM */
	// program an Atmel serial EEPROM with a bitstream
	bool ProgramEEPROM(istream& is);

	private:

	/* method: SetPROGRAM */
	// Set the value on the PROGRAM pin of the XC4000 FPGA
	void SetPROGRAM(unsigned int b);

	/* method: InitConfigureFPLD */
	// Initialize the XC4000 FPGA for configuration with DIN, CCLK, and PROGRAM pins
	void InitConfigureFPLD(void);

	/* method: ConfigureFPLD */
	// field type for the bitstream data
	// process a stream and send the configuration bitstream to the board
	bool ConfigureFPLD(istream& is);

	/* method: SetChipType */
	// field type for the FPGA device identifier
	// gets the chip identifier from the bitstream and stores it in the object
	bool SetChipType(istream& is);

	/* method: SendEEPROMByte */
	// Send a byte into an Atmel EEPROM
	void SendEEPROMByte(unsigned char byte);

	/* method: ProgramEEPROMPage */
	// Program a page of an Atmel EEPROM
	void ProgramEEPROMPage(unsigned char* buf, // buffer with data bytes
		unsigned int pageAddr, // address where data will be stored
		unsigned int len // number of bytes in buffer
		);

	/* method: SetChipType */
	// gets the chip identifier from the bitstream and stores it in the object
	bool SetChipType(string& bitfileName);

	/* end XC40 methods */

	string ramLoaderDir;	// directory where RAM interface bitstreams are found
};

/* begin XC40 functions */

/* function: GetInteger */
// compute an integer from a vector of hex numbers in a stream
static long unsigned int GetInteger(istream& is, int len=2);

/* function: GetType */
// get the field type indicator from a stream
static int GetType(istream& is);

/* function: ScanForField */
// pass over fields until a certain field is found
static bool ScanForField(istream& is, unsigned char searchType);

/* end XC40 functions */

#endif
