//
// The XC95 object specializes the XCPORT object so that the XC9500 CPLD on the
// XS95 Board can be configured and the RAM can be loaded.  The XC95 object controls
// the JTAG port of the XC9500 chip.
//

// $Log: xc95.cpp $
// Revision 1.2  1999/03/30 18:24:13  DEVB
// Fixed ConfigureFPLD routine so that it terminates with a true value
// upon the end of an SVF file.
//
// Revision 1.1  1999/03/29 22:03:30  DEVB
// Initial revision
//

#include <string.h>
#include <fstream.h>
#include <time.h>
#include <stdio.h>
#include "hex.h"
#include "bitstrm.h"
#include "xc95.h"
using std::string;

// length of JTAG boundary scan instruction register in XC9572,XC95108 CPLDs
const unsigned int bsirLength[] = {8,8};
// length of JTAG boundary scan data register in XC9572 CPLD
const unsigned int bsdrLength[] = {216,324};

// position of cells in the BSDR which affect the address bits of the
// RAM on the XS95 board
static unsigned int xc9500RamAddressBitPos[][numRAMAddressBits] =
{
	// A0 ... A14 for the XC9572
	{132,126,114,111,210,198,117,201,18,12,36,33,21,24,15},
	// A0 ... A14 for the XC95108
	{252,237,225,219,318,309,222,315,153,3,12,9,18,156,147},
};

// position of the cells in the BSDR which affect the data bits of the
// RAM on the XS95 board
static unsigned int xc9500RamDataBitPos[][numRAMDataBits] =
{
	// D0 ... D7 for the XC9572
	{48,69,72,78,57,63,66,87},
	// D0 ... D7 for the XC95108
	{57,63,66,72,75,81,84,90},
};

// position of the cells in the BSDR which affect the CS, OE, and WE
// bits of the RAM on the XS95 board
static unsigned int xc9500RamControlBitPos[][numRAMControlBits] =
{
	// CS, OE, WE for the XC9572
	{  9,  6,  159},
	// CS, OE, WE for the XC95108
	{135, 144, 138},
};



// begin methods and functions

// instantiate an object for loading an XS95 board on the parallel port
XC95::XC95(XSError* err, unsigned int portNum) : XCPort(err,portNum)
{
	ramLoaderIsRdy = false;	// boundary scan register descriptions are not yet initialized
}

// Initialize the XC9500 CPLD for configuration through the JTAG interface
void XC95::InitConfigureFPLD(void)
{
	InitTAP();	// initialize TAP controller of XC95 chip to reset state
}

// configure the CPLD so it can interface to the XS Board RAM
bool XC95::ConfigureForRAMDownload()
{
	// query the XC9500 chip identifier through the JTAG port
	if(DetectXC9500Type() == false)
	{
		ramLoaderIsRdy = false;
		return false;
	}
	int xc9500Type;
	if(chipType.substr(0,5) == "95108")
		xc9500Type = XC95108Type;
	else if(chipType.substr(0,4) == "9572")
		xc9500Type = XC9572Type;
	else
	{
		assert(1==0);	// should never get here!
	}

	// create the JTAG registers that will interface to the XS95 RAM
	bsirPtr = new Bitstream(bsirLength[xc9500Type]);
	bsdrPtr = new XCBSDR(bsdrLength[xc9500Type],
		xc9500RamAddressBitPos[xc9500Type],
		xc9500RamDataBitPos[xc9500Type],
		xc9500RamControlBitPos[xc9500Type]);
	ramLoaderIsRdy = true;	// now we are ready to download to the XS Board RAM
	return true;
}

// program an Atmel serial EEPROM with a bitstream
bool XC95::ProgramEEPROM(istream& is)
{
	return false; // there is no serial EEPROM on the XS95 board
}

// Send out a byte of configuration information to the XC9500 Flash
void XC95::ConfigureFPLD(unsigned char b)
{
	// JTAG configuration bytes contain values for both TDI and TMS,
	// so only four clock pulses are needed per byte
	for(int i=0; i<4; i++)
	{
		SetTDI(b&1);	// byte is sent out LSB first
		b >>= 1;
		SetTMS(b&1);
		b >>= 1;
		PulseTCK();
	}
}

static string NextWord(string& s)
{
	int pos = s.find_first_not_of(" \t");	// find beginning of word
	if(pos== -1)
	{	// no non-whitespace chars found, so string has no words
		s = "";
		return "";
	}
	s = s.substr(pos,s.length()-pos);	// remove leading whitespace
	pos = s.find_first_of(" \t");		// find end of first keyword
	string word;
	if(pos== -1)
	{	// no more whitespace, so entire string is the last word
		word = s;
		s = "";
	}
	else
	{	// extract the word from the string
		word = s.substr(0,pos);	// get first keyword from command
		s = s.substr(pos,s.length()-pos);	// remove keyword from command
	}
	return word;
}

// Process SVF and send results through JTAG configuration pins
bool XC95::ConfigureFPLD(istream& is)
{
	InitConfigureFPLD();
	GoThruTAPSequence(RunTestIdle,-1); // home state for configuration

	unsigned int seconds = clock()/CLOCKS_PER_SEC;

#if DEBUG_SVF
	long start = clock();
#endif
	
	int lineCnt = 0;
	unsigned long elapsedTime = 0;
	while(true)
	{
		const unsigned int maxLineLength = 400;
		char line[maxLineLength];
		line[0] = 0;
		is.getline(line,400);
		if(is.fail())
			break;
		lineCnt++;
		
		// indicate progress to user
		if((unsigned int)(clock()/CLOCKS_PER_SEC)!=seconds)
		{
			seconds = clock()/CLOCKS_PER_SEC;
			cerr << "+";
		}
		
		string cmd = line;	// store SVF command in a string variable
		
		// remove anything after the first ";" or "//" in the command
		unsigned int pos = cmd.find_first_of(';');
		cmd = cmd.substr(0,pos);
		pos = cmd.find("//");
		cmd = cmd.substr(0,pos);
		
#if DEBUG_SVF
		cout << cmd.data() << endl; cout.flush();
#endif
		
		if(cmd.length()==0 || cmd.compare("")==0)
			continue;	// must be a comment line
		
		// remove the first keyword from the command
		string keywd = NextWord(cmd);
		if(keywd == "")
			continue;	// must be a blank line

		int status;
		unsigned long len;
		static Bitstream hdr(0), hir(0), tdr(0), tir(0);
		
		if(keywd.compare("TRST")==0)
		{ // TRST command has nothing to do with configuration so ignore it
#if DEBUG_SVF
			cout << "TRST" << endl; cout.flush();
#endif
			continue;
		}
		
		else if(keywd.compare("STATE")==0)
		{ // STATE command has nothing to do with configuration so ignore it
#if DEBUG_SVF
			cout << "STATE" << endl; cout.flush();
#endif
			continue;
		}
		
		else if(keywd.compare("ENDIR")==0)
		{ // ENDIR has nothing to do with configuration, but check to make
			// sure it is a well-formed command
#if DEBUG_SVF
			cout << "ENDIR" << endl; cout.flush();
#endif
			keywd = NextWord(cmd);
			assert(keywd.compare("IDLE")==0);	// make sure it's what should be there
			continue;
		}
		
		else if(keywd.compare("ENDDR")==0)
		{ // ENDDR has nothing to do with configuration, but check to make
			// sure it is a well-formed command
#if DEBUG_SVF
			cout << "ENDDR" << endl; cout.flush();
#endif
			keywd = NextWord(cmd);
			assert(keywd.compare("IDLE")==0);	// make sure it's what should be there
			continue;
		}
		
		else if(keywd.compare("HIR")==0)
		{ // initialize the header bits for loading into the IR
#if DEBUG_SVF
			cout << "HIR" << endl; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			Bitstream bs(len);
			unsigned int bsi = 0; // bit stream index
			// process hex number within parentheses
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			hir = bs; // copy bitstream into IR header bitstream
			continue;
		}
		
		else if(keywd.compare("HDR")==0)
		{ // initialize the header bits for loading into the DR
#if DEBUG_SVF
			cout << "HDR" << endl; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			Bitstream bs(len);
			unsigned int bsi = 0; // bit stream index
			// process hex number within parentheses
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			hdr = bs; // copy bitstream into IR header bitstream
			continue;
		}
		
		else if(keywd.compare("TIR")==0)
		{ // initialize the trailer bits for loading into the IR
#if DEBUG_SVF
			cout << "TIR" << endl; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			Bitstream bs(len);
			unsigned int bsi = 0; // bit stream index
			// process hex number within parentheses
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			tir = bs; // copy bitstream into IR header bitstream
			continue;
		}
		
		else if(keywd.compare("TDR")==0)
		{ // initialize the trailer bits for loading into the DR
#if DEBUG_SVF
			cout << "TDR" << endl; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			Bitstream bs(len);
			unsigned int bsi = 0; // bit stream index
			// process hex number within parentheses
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			tdr = bs; // copy bitstream into IR header bitstream
			continue;
		}
		
		else if(keywd.compare("SIR")==0)
		{ // send a bitstream into the instruction register
#if DEBUG_SVF
			cout << "SIR" << endl; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			// create a bitstream
			Bitstream bs(len);
			unsigned int bsi=0;
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			// concatenate header, data, and trailer bitstreams
			bs = hir + bs + tir;
#if DEBUG_SVF
			cout << bs << endl; cout.flush();
#endif
			// send bitstream into IR and return to Run-Test-Idle state
			GoThruTAPSequence(SelectDRScan,SelectIRScan,CaptureIR,ShiftIR,-1);
			SendRcvBitstream(bs,bs);
			GoThruTAPSequence(PauseIR,Exit2IR,UpdateIR,RunTestIdle,-1);
			continue;
		}
		
		else if(keywd.compare("SDR")==0)
		{ // send a bitstream into the data register
#if DEBUG_SVF
			cout << "SDR" << endl; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			// create a bitstream
			Bitstream bs(len);
			unsigned int bsi=0;
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			// concatenate header, data, and trailer bitstreams
			bs = hdr + bs + tdr;
#if DEBUG_SVF
			cout << bs << endl; cout.flush();
#endif
			// send bitstream into IR and return to Run-Test-Idle state
			GoThruTAPSequence(SelectDRScan,CaptureDR,ShiftDR,-1);
			SendRcvBitstream(bs,bs);
			GoThruTAPSequence(PauseDR,Exit2DR,UpdateDR,RunTestIdle,-1);
			continue;
		}
		
		else if(keywd.compare("RUNTEST")==0)
		{ // insert delay while programming takes place
#if DEBUG_SVF
			cout << "RUNTEST" << endl;
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			elapsedTime += len;
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TCK")==0);
#if DEBUG_SVF
			cout << "delay = " << len << " us" << endl; cout.flush();
#endif
			InsertDelay(len,MICROSECONDS);
			continue;
		}
		
		// if we get here, then it's some strange junk we don't understand
		XSError& err = GetErr();
		err.SetSeverity(XSErrorFatal);
		err << "unknown SVF command: " << cmd.data() << endl;
		err.EndMsg();
		return false;
	}
	
	cerr << endl;
	
#if DEBUG_SVF
	long finish = clock();
	cerr << "elapsed time = " << (finish-start)/(CLOCKS_PER_SEC) << endl;
	cerr << "elapsed time = " << elapsedTime << " us\n";
#endif
	
	InitTAP(); // done so reset TAP and idle the JTAG logic
	return true;
}

// detect the presence and type of the XC9500 CPLD chip on the port
// and set it into the chip identifier field of the object
bool XC95::DetectXC9500Type()
{
	Bitstream bsir(8);  // maximum BSIR
	Bitstream bsdr(488);  // maximum BSDR
	bsir.SetBits(0, 0,1,1,1,1,1,1,1, -1); // IDCODE instruction for reading chip id
	
	Bitstream null(0);	// zero-length bitstream

	// identifiers for the various types of XC9500 CPLD chips in 84-pin PLCC
	Bitstream XC9572ID(32), XC95108ID1(32), XC95108ID2(32);
	XC9572ID.SetBits( 0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,0,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1,0,0,0,0, -1);
	XC95108ID1.SetBits(0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,1,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1,0,0,0,0, -1);
	XC95108ID2.SetBits(0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,1,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1,1,0,0,0, -1);
	
	InitTAP(); // initialize the JTAG TAP controller
	LoadBSIRthenBSDR(bsir,null,bsdr); // get the ID codes from the chips
	
	if(XC9572ID.Subcompare(0,bsdr,0)==1)
	{ // chip id matches XC9572
		SetChipType((string)"XC9572PC84");
		return true;
	}
	else if(XC95108ID1.Subcompare(0,bsdr,0)==1)
	{ // chip id matches XC95108
		SetChipType((string)"XC95108PC84");
		return true;
	}
	else if(XC95108ID2.Subcompare(0,bsdr,0)==1)
	{ // chip id matches XC95108
		SetChipType((string)"XC95108PC84");
		return true;
	}
	else
	{ // some non-XC9500 chip is on the port
		SetChipType((string)"UNKNOWN");
		return false;
	}
	return false;
}
