/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// XCPort objects specialize a JTAGPort to handle programming of
// XS40 and XS95 Boards.
//

// $Log: xcport.cpp $
// Revision 1.3  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.2  1999/03/30 15:29:10  DEVB
// Changed InsertDelay function so that it has less chance of overflowing.
// Also disabled optimization from removing the empty timing loops.
//
// Revision 1.1  1999/03/29 21:59:44  DEVB
// Initial revision
//

//#include <dos.h>
#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#ifdef __WIN32__
#include <strstrea.h>
#else
#include <strstream.h>
#endif
#include <time.h>
#include <string.h>

#include "xcport.h"
#include "utils.h"

// XCPort controls the JTAG port of the XC4000 and XC9500 chips, and also
// the XC4000 chip via its /PROGRAM, DIN, and CCLK pins.  The download/JTAG
// pins for the XS95 and XS40 boards are connected to the control outputs
// of the printer port (write address offset = 2) and the status inputs
// of the printer port (read address offset = 1).  The inversion mask for
// the outputs is:
//		1  C0-  drives /PROGRAM pin of the XC4000 directly,
//		0  C1-  drives TCK and CCLK through an inverter on the board,
//		0  C2+  drives TMS directly,
//		1  C3-  drives TDI and DIN directly.
// The inversion mask for the inputs is:
//		X  S0	unused
//		X  S1	unused
//		X  S2	unused
//		0  S3+  reads P15 of microcontroller directly,
//		0  S4+  reads P14 of microcontroller directly,
//		0  S5+  reads P16 of microcontroller directly,
//		0  S6+	reads P31 of microcontroller directly,
//		0  S7-  reads TDO/DOUT through an inverter.

// begin methods and functions

// instantiate an XCPort object on a given parallel port
XCPort::XCPort(XSError* e, unsigned int portNum) : JTAGPort(e,portNum,0x03,0,0x09)
{
	chipType = "UNKNOWN";
	ramLoaderIsRdy = false;	// interface between FPLD and RAM does not start in ready state
}

// sets the chip identifier in the object
bool XCPort::SetChipType(const string& type)
{
	int pos = type.find_first_of("0123456789");
	if(pos == -1)
	{	// no numbers!? not a xilinx chip id!
		chipType = "UNKNOWN";
		return false;
	}
	// remove any beginning non-numeric characters from chip type
	chipType = type.substr(pos,type.length()-pos);	
	int len = chipType.length();
	len = len>7 ? 7:len;	// keep only the first 7 characters
	chipType = chipType.substr(0,len);
	ConvertToUpperCase(chipType);
	return true;
}

// download the XS Board RAM with the contents arriving through a stream
bool XCPort::DownloadRAM(istream& is)
{
	XSError& err = GetErr();

	if(ramLoaderIsRdy==false)
		if(ConfigureForRAMDownload()==false)
			return false;

	XCBSDR& bsdr = *bsdrPtr;
	Bitstream& bsir = *bsirPtr;
	Bitstream null(0);
	// get to a state where the controller will accept an instruction
	GoThruTAPSequence(RunTestIdle,SelectDRScan,SelectIRScan,CaptureIR,ShiftIR, -1);
	// first load an EXTEST instruction into the BSIR
	bsir.Clear();	// all-zeroes is the EXTEST opcode
	SendRcvBitstream(bsir, null);  // after this, TAP state is Exit1-IR
	// activate the EXTEST instruction and get ready to send data to the DR
	GoThruTAPSequence(UpdateIR, SelectDRScan, CaptureDR, ShiftDR, -1);
	
	// enable the RAM
	bsdr.Clear();
	bsdr.SetRAMControls(0,1,1); // enable RAM, disable writing
	SendRcvBitstream(bsdr,null);
	GoThruTAPSequence(UpdateDR,SelectDRScan,CaptureDR,ShiftDR,-1);
	
	// read hex records from file and place data in the RAM of the board
	while(is.eof()==0)
	{
		HexRecord hx;
		is >> hx;
		if(is.eof()!=0)
			break; // terminate loading loop when stream goes dry
		if(is.fail() || err.IsError())
		{
			err.SetSeverity(XSErrorFatal);
			err << "error reading stream" << endl;
			err.EndMsg();
			break;
		}
		if(hx.IsError())
		{ // some error in the hex record itself
			err.SetSeverity(XSErrorMajor);
			err << "hex record: " << hx.GetErrMsg() << endl;
			err.EndMsg();
			break;
		}
		else
		{ // send out an indication for each hex record that's loaded
			cerr << ".";
			DownloadHexRecordToRAM(hx);
		}
	}
	cerr << endl;
	
	bsdr.SetRAMControls(1,1,1); // disable RAM
	SendRcvBitstream(bsdr,null);
	GoThruTAPSequence(UpdateDR,-1);
	
	InitTAP();  // reset the TAP state controller
	
	return true;
}

// load RAM data at the addresses given in a hex record
// (This routine assumes the entering TAP state is Shift-DR.
// It exits with the TAP state being Shift-DR.)
void XCPort::DownloadHexRecordToRAM(HexRecord& hx)
{
	Bitstream null(0);
	if(hx.IsData()) // don't download unless this is RAM data
	{
		// process the bytes from a data-type hex record
		for(unsigned int i=0; i<hx.GetLength(); i++)
		{
			XCBSDR& bsdr = *bsdrPtr;
			
			// load address and data onto the FPLD pins
			bsdr.Clear();	// set all the BSDR bits to zero
			bsdr.SetRAMAddress(hx.GetAddress()+i);
			bsdr.SetRAMData(hx[i]);
			bsdr.SetRAMControls(0,1,1); // enable RAM, disable outputs, raise write line
			SendRcvBitstream(bsdr,null);
			GoThruTAPSequence(UpdateDR,SelectDRScan,CaptureDR,ShiftDR,-1);
			
			bsdr.SetRAMControls(0,1,0); // bring write enable low
			SendRcvBitstream(bsdr,null);
			GoThruTAPSequence(UpdateDR,SelectDRScan,CaptureDR,ShiftDR,-1);
			
			bsdr.SetRAMControls(0,1,1); // bring write enable high again
			SendRcvBitstream(bsdr,null);
			GoThruTAPSequence(UpdateDR,SelectDRScan,CaptureDR,ShiftDR,-1);
		}
	}
}

// get RAM data at the addresses given and store it in a hex record
// (This routine assumes the entering TAP state is Shift-DR.
// It exits with the TAP state being Shift-DR.)
void XCPort::UploadHexRecordFromRAM(
						HexRecord& hx,			// hex record to store data in
						unsigned long loAddr,	// begin upload from this address
						unsigned long hiAddr	// end upload at this address
						)
{
	XSError& err = GetErr();
	
	hx.SetTag(DataRecord);
	hx.SetAddress(loAddr);
	hx.SetLength(hiAddr-loAddr);
	
	XCBSDR bsdrOut = *bsdrPtr;
	XCBSDR bsdrIn = *bsdrPtr;
	Bitstream null(0);
	
	// process the bytes from a data-type hex record
	for(unsigned long i=loAddr; i<hiAddr; i++)
	{
		// load address and data onto the FPLD pins
		bsdrOut.Clear();
		bsdrOut.SetRAMAddress(i);
		bsdrOut.ReadRAMData();  // tristate FPLD data lines so RAM has control
		bsdrOut.SetRAMControls(0,0,1); // enable RAM, enable outputs, keep write high
		SendRcvBitstream(bsdrOut,null);
		GoThruTAPSequence(UpdateDR,SelectDRScan,CaptureDR,ShiftDR,-1);
		
		bsdrOut.SetRAMControls(0,1,1); // disable RAM outputs
		SendRcvBitstream(bsdrOut,bsdrIn); // shift out RAM data while loading new data
		unsigned long drivenAddress = bsdrIn.GetRAMAddress();
		if(drivenAddress != i)
		{
			err.SetSeverity(XSErrorMinor);
			err << "specified memory address = " << i
				<< " but actual address = " << drivenAddress << endl;
			err.EndMsg();
			hx[i-loAddr] = 0xff;
		}
		else
			hx[i-loAddr] = bsdrIn.GetRAMData();
		
		GoThruTAPSequence(UpdateDR,SelectDRScan,CaptureDR,ShiftDR,-1);
	}
	hx.CalcCheckSum();
}

// upload the XS Board RAM to a stream
bool XCPort::UploadRAM(
					   ostream& os,				// dump uploaded data to this stream
					   unsigned long loAddr,	// start fetching data from this address
					   unsigned long hiAddr		// stop at this address
					   )
{
	XSError& err = GetErr();
	
	XCBSDR& bsdr = *bsdrPtr;
	Bitstream& bsir = *bsirPtr;
	Bitstream null(0);
	// get to a state where the TAP controller will accept an instruction
	GoThruTAPSequence(RunTestIdle,SelectDRScan,SelectIRScan,CaptureIR,ShiftIR, -1);
	// first load an EXTEST instruction into the BSIR
	bsir.Clear();	// all-zeroes is the EXTEST opcode
	SendRcvBitstream(bsir, null);  // after this, TAP state is Exit1-IR
	// activate the EXTEST instruction and get ready to send data to the DR
	GoThruTAPSequence(UpdateIR, SelectDRScan, CaptureDR, ShiftDR, -1);
	
	// enable the RAM so data can be read out of it
	bsdr.Clear();
	bsdr.SetRAMControls(0,1,1); // enable RAM, disable writing
	SendRcvBitstream(bsdr,null);
	// transfer TAP to ShiftDR state so RAM data can be extracted via JTAG
	GoThruTAPSequence(UpdateDR,SelectDRScan,CaptureDR,ShiftDR,-1);
	
	// read hex records from RAM and place data in the hex file
	for(unsigned long addr=loAddr&~0xF; addr<hiAddr; addr+=16)
	{
		HexRecord hx;
		UploadHexRecordFromRAM(hx,addr,addr+16); // get 16 bytes from RAM
		os << hx;		// send hex record to output stream
		cerr << ".";	// give feedback on progress
	}
	
	bsdr.SetRAMControls(1,1,1); // disable RAM
	SendRcvBitstream(bsdr,null);
	GoThruTAPSequence(UpdateDR,-1);
	
	InitTAP();  // reset the TAP state controller
	
	return true;
}

bool XCPort::UploadRAM(
					   string& hexfileName,		// dump uploaded data to this file
					   unsigned long loAddr,	// start fetching data from this address
					   unsigned long hiAddr		// stop at this address
					   )
{
	XSError& err = GetErr();
	
	if(hexfileName.length()==0)
		return false;  // stop if no hex file was given
	
	cerr << "Uploading hex records to " << hexfileName.data()
		<< " into XS Board RAM" << endl;
	
	ofstream os(hexfileName.data());  // otherwise open hex file
	if(os.fail() || os.eof()!=0)
	{ // error - couldn't open hex file
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << hexfileName.data() << endl;
		err.EndMsg();
	}
	
	UploadRAM(os,loAddr,hiAddr);
	
	os.close();  // close-up the hex file
	
	return true;
}
// download the XS Board RAM with the contents of a HEX file
bool XCPort::DownloadRAM(string& hexfileName)
{
	XSError& err = GetErr();
	
	if(hexfileName.length()==0)
		return false;  // stop if no hex file was given
	
	cerr << "Downloading hex records from " << hexfileName.data()
		<< " into XS Board RAM" << endl;
	
	ifstream is(hexfileName.data(),ios::nocreate);  // otherwise open hex file
	if(is.fail() || is.eof())
	{ // error - couldn't open hex file
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << hexfileName.data() << endl;
		err.EndMsg();
	}
	
	DownloadRAM(is);
	
	is.close();  // close-up the hex file
	
	return true;
}

// configure the FPLD with the bit stream stored in a BIT or SVF file
bool XCPort::ConfigureFPLD(string& fileName)
{
	XSError& err = GetErr();
	
	if(fileName.length()==0)
		return false;  // stop if no SVF or BIT file was given
	
	cerr << "Configuring FPLD with bitstream from " << fileName.data() << endl;
	
	ifstream is(fileName.data(),ios::binary|ios::nocreate);  // otherwise open BIT or SVF file
	if(is.fail() || is.eof())
	{ // error - couldn't open file
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << fileName.data() << endl;
		err.EndMsg();
	}
	
	bool status = ConfigureFPLD(is);
	
	is.close();  // close-up the hex file
	
	return status;
}

// program the Atmel serial EEPROM on the XS Board with a bitstream from a file
bool XCPort::ProgramEEPROM(string& bitfileName)
{
	XSError& err = GetErr();
	
	if(bitfileName.compare("")==0)
		return false; // exit if no BIT file was given
	
	ifstream is(bitfileName.data(),ios::binary|ios::nocreate);
	if(is.fail() || is.eof())
	{ // error if given file could not be opened
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << bitfileName.data() << endl;
		err.EndMsg();
	}
	
	// now configure the FPGA from the bitfile stream
	bool status = ProgramEEPROM(is);
	
	is.close();
	
	return status;
}


#pragma optimize( "", off ) // turn off optimization so timing loops don't get removed

// subroutine for delaying by a given number of microseconds or milliseconds
void XCPort::InsertDelay(unsigned long d, unsigned int time_units)
{
	static long loopsPerMillisecond = -1;
	
	if(loopsPerMillisecond < 0)
	{ // determine the timing of the computer if we haven't already done so
		clock_t start, finish;
		start = clock();	// get starting time
		for(long i = 10000000L; i>0L; i--)
			;	// do a whole bunch of empty loops
		finish = clock();	// get ending time
		assert(finish!=start);
		// compute the number of empty loops per millisecond
		loopsPerMillisecond = (10000L*CLOCKS_PER_SEC)/(finish-start);
		assert(loopsPerMillisecond>1000);	// we need enough precision
#if DEBUG
		cerr << "finish = " << finish << "   start = " << start << endl;
		cerr << "loops per ms = " << loopsPerMillisecond << endl;
#endif
	}
	
	if(time_units == MILLISECONDS)
	{	// doing millisecond timing here
		for(long loops = loopsPerMillisecond*d; loops>0; loops--)
			;
	}
	else if(time_units==MICROSECONDS && d>10000)
	{	// doing millisecond timing here but with a microsecond time base
		for(long loops = (loopsPerMillisecond*d)/1000; loops>0; loops--)
			;
	}
	else
	{	// doing microsecond timing here
		for(long loops = (loopsPerMillisecond*d)/1000; loops>0; loops--)
			;
	}
}

#pragma optimize( "", on )
