/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// XCPort objects specialize a JTAGPort to handle programming of
// XS40 and XS95 Boards.
//

// $Log: xcport.h $
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:50:24  DEVB
// Initial revision
//

#ifndef XCPort_H
#define XCPort_H

#include <fstream.h>

#include "xcbsdr.h"
#include "jtagport.h"
#include "hexrecrd.h"

const unsigned int MILLISECONDS = 1000;
const unsigned int MICROSECONDS = 1;

class XCPort : public JTAGPort
{
	public:

	/* begin XCPort methods */

	/* method: XCPort */
	// instantiate an XCPort object on a given parallel port
	 XCPort(XSError* e, unsigned int portNum);

	/* method: SetChipType */
	// sets the chip identifier in the object
	bool SetChipType(const string& type);

	/* method: DownloadRAM */
	// download the XS Board RAM with the contents arriving through a stream
	bool DownloadRAM(istream& is);
	
	// configure the FPLD so it can interface to the XS Board RAM
	virtual bool ConfigureForRAMDownload() = 0;

	/* method: UploadRAM */
	// upload the XS Board RAM to a stream
	bool UploadRAM(
		ostream& os, // dump uploaded data to this stream
		unsigned long loAddr, // start fetching data from this address
		unsigned long hiAddr // stop at this address
		);

	/* method: UploadRAM */
	bool UploadRAM(
		string& hexfileName, // dump uploaded data to this file
		unsigned long loAddr, // start fetching data from this address
		unsigned long hiAddr // stop at this address
		);

	/* method: DownloadRAM */
	// download the XS Board RAM with the contents of a HEX file
	bool DownloadRAM(string& hexfileName);

	/* method: ConfigureFPLD */
	// configure the FPLD with the bit stream stored in a BIT or SVF file
	bool ConfigureFPLD(string& fileName);
	
	// configure the FPLD with the bit stream arriving through a stream
	virtual bool ConfigureFPLD(istream& is) = 0;

	/* method: ProgramEEPROM */
	// program the Atmel serial EEPROM on the XS Board with a bitstream from a file
	bool ProgramEEPROM(string& bitfileName);

	// program EEPROM on the XS Board
	virtual bool ProgramEEPROM(istream& is) = 0;

	/* method: InsertDelay */
	// subroutine for delaying by a given number of microseconds
	void InsertDelay(unsigned long d, unsigned int time_units);
	
	protected:
	
	// The RAM of the XS board is controlled through a JTAG interface,
	string chipType;		// type of chip (e.g. 4005XLPC84)
	XCBSDR* bsdrPtr;		// JTAG data register useful for XS RAM access
	Bitstream* bsirPtr;		// JTAG instruction register
	bool ramLoaderIsRdy;	// true if the interface to download to the XS board RAM is ready

	private:

	/* method: DownloadHexRecordToRAM */
	// load RAM data at the addresses given in a hex record
	// (This routine assumes the entering TAP state is Shift-DR.
	// It exits with the TAP state being Shift-DR.)
	void DownloadHexRecordToRAM(HexRecord& hx);

	/* method: UploadHexRecordFromRAM */
	// get RAM data at the addresses given and store it in a hex record
	// (This routine assumes the entering TAP state is Shift-DR.
	// It exits with the TAP state being Shift-DR.)
	void UploadHexRecordFromRAM(
		HexRecord& hx, // hex record to store data in
		unsigned long loAddr, // begin upload from this address
		unsigned long hiAddr // end upload at this address
		);

	/* end XCPort methods */

};

/* begin XCPort functions */

/* end XCPort functions */

#endif
