/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XSLOAD program has the following functions:
//		1. Programs XS40 Boards with XC4000 BIT files
//		2. Programs XS95 Boards with XC9500 SVF files
//		3. Loads the XS Board RAM with HEX data files
//		4. Programs the Serial EEPROM on the XS40 Board
//

// $Log: xsload.cpp $
// Revision 1.3  1999/03/30 19:46:28  DEVB
// Added code to display release version number
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 21:49:31  DEVB
// Initial revision
//

#include <stdlib.h>
#include <assert.h>
#ifdef __WIN32__
#include <strstrea.h>
#else
#include <strstream.h>
#endif
#include <fstream.h>
#include <string>
using std::string;

#include "xc40.h"
#include "xc95.h"
#include "hex.h"
#include "utils.h"

//
// global variables
//
string version("$Name: V2_0_0 $");
XSError errMsg(cerr); // setup error channel
unsigned int portNum = 1; // default parallel port number
string loadFileName = "";	// name of 32 KByte RAM loading circuit
string hexFileName  = "";	// name of file to be loaded into RAM (.HEX)
string svfFileName  = "";	// name of run-time circuit (.SVF) for XC9500
string bitFileName  = "";	// name of run-time circuit (.BIT) for XC4000
string dumpFileName = "";	// name of RAM data dump file
unsigned int numHexFiles     = 0;	// number of hex data files on the command line
unsigned int numSvfFiles     = 0;	// number of XC9500 SVF files on the command line
unsigned int numBitFiles     = 0;	// number of XC4000 BIT files on the command line
string xstoolsBinDir = "";	// directory containing this program executable
bool programSerialEEPROM = false;

//
// print a string that shows the options and how to use this program.
//
void Usage(const string& progName)
{
	cerr << progName.data() << " [-[h|help]] [-[p|port] #[1,2,3]] [-serial_eeprom] file1[.HEX] file2[.BIT|.SVF]" << endl;
	int pos = version.find("Name:");
	if(pos < 0)
	{
		cerr << "Version number unavailable!!" << endl;
	}
	else
	{
		string versionNumber = version.substr(pos+5,version.length()-pos-5);
		for(int i=versionNumber.length()-1; i>=0; i--)
		{
			if(versionNumber[i]=='_')
				versionNumber[i] = '.';
			else if(versionNumber[i]=='$')
				versionNumber[i] = ' ';
		}
		cerr << "Version " << versionNumber.data() << endl;
	}
}

//
// process the command line options and set the appropriate flags.
//
void ProcessOpts(int* argc, char** argv)
{
	int i;
	for(i=1; i< *argc; i++)
	{
		string arg = argv[i];
		
		if(arg[0] == '-')  // found an option
		{
			int cnt;
			ConvertToUpperCase(arg);	// convert option to upper case
			
			// ask for help but we don't give much
			if(arg=="-H" || arg=="-HELP")
			{
				Usage(arg);
				cnt = 1;
			}
			
			// specify the printer port being used
			else if(arg=="-PORT" || arg=="-P")
			{
				istrstream argstr(argv[i+1]);
				argstr >> portNum;
				cnt = 2;
			}
			
			// specify that the serial EEPROM is to be programmed
			else if(arg=="-SERIAL_EEPROM" || arg=="-SEEP")
			{
				programSerialEEPROM = true;
				cnt = 1;
			}
			
			// can't figure out what the user wants
			else
			{
				errMsg.SetSeverity(XSErrorMinor);
				errMsg << "unknown option: " << arg.data() << endl;
				errMsg.EndMsg();
				Usage(string(argv[0]));
				cnt = 1;
			}
			
			// remove the processed options from the argument list
			for(int j=i+cnt; j< *argc; j++)
				argv[j-cnt] = argv[j];
			*argc -= cnt;
			i--;
		}
	}
	
	// all the options are processed, now handle the download files from
	// the command line
	string suffixes[] = {"SVF","BIT","HEX"};
	int numFiles[3] = {0,};		// records number of each file type in command line
	string fileName[3];			// file names for SVF, BIT, and HEX files
	for(i=1; i<*argc; i++)
	{
		ifstream fs;
		string arg = argv[i];
		string suffix = GetSuffix(arg);
		ConvertToUpperCase(suffix);
		if(suffix.empty()==true)
		{ // no suffix, so see what kinds of files are present
			bool found = false;
			for(int j=0; j<3 && found==false; j++)
			{
				string tmpFileName = arg + "." + suffixes[j];
				fs.open(tmpFileName.data(),ios::nocreate);
				if(fs.good())
				{ // yes, a file with this suffix and base name exists
					found = true;
					numFiles[j]++;
					fileName[j] = tmpFileName;
					fs.close();
				}
			}
			if(found==false)
			{ // no file with any of the allowed suffixes was found
				errMsg.SetSeverity(XSErrorMinor);
				errMsg << "can't find .BIT, .SVF, or .HEX file for this: "
					<< arg.data() << endl;
				errMsg.EndMsg();
				Usage(string(argv[0]));
				return;
			}
		}
		else
		{ // suffix found, so see what the file suffix is
			int j;
			for(j=0; j<3; j++)
			{
				if(suffix == suffixes[j])
				{
					numFiles[j]++;
					fileName[j] = argv[i];
					break;
				}
			}
			if(j>=3)
			{
				errMsg.SetSeverity(XSErrorMinor);
				errMsg << "unknown file type: " << arg.data() << endl;
				errMsg.EndMsg();
				Usage(string(argv[0]));
				return;
			}
		}
	}
	numSvfFiles += numFiles[0];
	numBitFiles += numFiles[1];
	numHexFiles += numFiles[2];
	svfFileName = fileName[0];
	bitFileName = fileName[1];
	hexFileName = fileName[2];
	
	if(numHexFiles>1)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "only one HEX file is allowed" << endl;
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}
	if(numBitFiles>1)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "only one BIT file is allowed" << endl;
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}
	if(numSvfFiles>1)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "only one SVF file is allowed" << endl;
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}
	if(numBitFiles==0 && numSvfFiles==0  && numHexFiles==0)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "at least one BIT, SVF, or HEX file is needed" << endl;
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}

	char* dir = getenv("XSTOOLS_BIN_DIR");
	if(dir==NULL)
	{
		errMsg.SetSeverity(XSErrorFatal);
		errMsg << "XSTOOLS_BIN_DIR environment variable is not set" << endl;
		errMsg.EndMsg();
	}
	xstoolsBinDir = dir;
}


main( int argc, char **argv )
{
	// set the error message header to the name of this program
	string cmd = argv[0];
	cmd = StripSuffix(StripPrefix(cmd));
	errMsg.SetHeader(cmd);
	
	ProcessOpts( &argc, argv );  // process command line options
	if(errMsg.IsError()) // exit if errors in options
		exit(1);
	
	XCPort* portPtr;
	if(numBitFiles>0)
		portPtr = new XC40(&errMsg,portNum,bitFileName,xstoolsBinDir);
	else
		portPtr = new XC95(&errMsg,portNum);
	XCPort& port = *portPtr;
	
	if(numHexFiles>0)
	{
		port.DownloadRAM(hexFileName);
		if(errMsg.IsError())
			exit(1);
	}
	
	if(numBitFiles>0)
	{
		if(programSerialEEPROM==true)
			port.ProgramEEPROM(bitFileName);
		else
			port.ConfigureFPLD(bitFileName);
		if(errMsg.IsError())
			exit(1);
	}
	else if(numSvfFiles>0)
	{
		port.ConfigureFPLD(svfFileName);
		if(errMsg.IsError())
			exit(1);
	}
	exit(0);
}
